-------------------------------------------------------------------------------
--! @file       CipherCore_TP_Control.vhd
--! @brief      Control unit for AEZ
--! @project    CAESAR Candidate Evaluation
--! @author     Ekawat (ice) Homsirikamol
--! @copyright  Copyright (c) 2016 Cryptographic Engineering Research Group
--!             ECE Department, George Mason University Fairfax, VA, U.S.A.
--!             All rights Reserved.
--! @license    This project is released under the GNU Public License.
--!             The license and distribution terms for this file may be
--!             found in the file LICENSE in this distribution or at
--!             http://www.gnu.org/licenses/gpl-3.0.txt
--! @note       This is publicly available encryption source code that falls
--!             under the License Exception TSU (Technology and software-
--!             —unrestricted)
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity CipherCore_TP_Control is
    port (
        clk              : in  std_logic;
        rst              : in  std_logic;

        --! Input        
        key_update       : in  std_logic;
        bdi_valid        : in  std_logic;
        bdi_ready        : out std_logic;
        decrypt          : in  std_logic;
        len_a            : in  std_logic_vector(16           -1 downto 0);
        len_d            : in  std_logic_vector(16           -1 downto 0);

        --! Datapath
        init_key_done    : in  std_logic;
        ctrl             : out std_logic_vector(38           -1 downto 0);
        ctrl_e           : out std_logic_vector(17           -1 downto 0);
        rtr              : out std_logic_vector( 4           -1 downto 0);
        data_size        : out std_logic_vector( 4           -1 downto 0);
        edone            : in  std_logic;
        eready           : in  std_logic;

        --! FIFO
        fdi_ready        : out std_logic;
        fdo_valid        : out std_logic;
        fdo_ready        : in  std_logic;
        fdi_valid        : in  std_logic;

        --! Output
        msg_auth_done    : out std_logic;
        bdo_valid        : out std_logic;
        bdo_ready        : in  std_logic        
    );
end entity CipherCore_TP_Control;

architecture behavior of CipherCore_TP_Control is
    constant TINY_RND           : integer := 8;
    --! Top
    signal en_out               : std_logic;
    signal en_ll                : std_logic;
    signal en_xy                : std_logic;
    signal en_xy2               : std_logic;
    signal en_hash              : std_logic;
    signal en_ss                : std_logic;
    signal en_tmp               : std_logic;
    signal sel_mask             : std_logic;
    signal sel_pad              : std_logic;
    signal sel_tau              : std_logic;
    signal sel_rot              : std_logic_vector(2            -1 downto 0);
    signal sel_ll               : std_logic;
    signal sel_rothb            : std_logic;
    signal sel_ldata            : std_logic_vector(3            -1 downto 0);
    signal sel_ldata_xor        : std_logic_vector(2            -1 downto 0);
    signal sel_data             : std_logic_vector(3            -1 downto 0);
    signal sel_tiny             : std_logic;
    signal sel_xy               : std_logic_vector(2            -1 downto 0);
    signal sel_tag              : std_logic;
    signal sel_output           : std_logic;
    signal sel_fifo             : std_logic;
    signal tinyround            : std_logic_vector(3            -1 downto 0);
    signal sel_sout             : std_logic;
    signal sel_tmp_xor          : std_logic;
    signal sel_e                : std_logic;
    signal sel_otag             : std_logic;
    signal en_tag               : std_logic;
    --!     AEZ_E
    signal eclr                 : std_logic;
    signal estart               : std_logic;
    signal einit_key            : std_logic;
    signal dcalc_mode           : std_logic_vector(3            -1 downto 0);
    signal dcalc_init           : std_logic_vector(3            -1 downto 0);
    signal block_number         : std_logic_vector(8            -1 downto 0);


    --! Internal signals
    --!     Registers
    signal is_decrypt           : std_logic;
    signal len_a_r              : std_logic_vector(16           -1 downto 0);
    signal len_d_r              : std_logic_vector(16           -1 downto 0);
    signal len_d_orig_r         : std_logic_vector(16           -1 downto 0);
    signal is_odd               : std_logic;
    signal is_secondpass        : std_logic;
    signal en_xy_r              : std_logic;
    signal edone_r              : std_logic;
    signal set_save_hash_r      : std_logic;
    signal is_postproc          : std_logic;
    signal is_notfirst          : std_logic;
    --!     Set register control & register
    signal set_secondpass       : std_logic;
    signal set_store_sum        : std_logic;
    signal clr_store_sum        : std_logic;
    signal store_sum            : std_logic;
    signal set_save_hash        : std_logic;
    signal save_hash            : std_logic;
    signal set_pad              : std_logic;
    signal clr_pad              : std_logic;
    signal is_pad               : std_logic;
    --!     To external
    signal bdi_ready_s          : std_logic;
    signal bdo_valid_s          : std_logic;
    signal msg_auth_done_s      : std_logic;
    --!     Signals
    signal clr_init             : std_logic;
    signal en_tinyround         : std_logic;
    signal clr_tinyround        : std_logic;
    signal clr_sel_ll           : std_logic;
    signal set_sel_ll           : std_logic;
    signal set_postproc         : std_logic;
    signal set_notfirst         : std_logic;

    signal sel_i                : std_logic;
    signal set_odd              : std_logic;
    signal new_i                : std_logic_vector(3            -1 downto 0);
    signal rtr_s                : std_logic_vector(4            -1 downto 0);
    signal masksize             : std_logic_vector(4            -1 downto 0);
    signal sel_dsize            : std_logic_vector(3            -1 downto 0);
    signal sel_rot_size         : std_logic_vector(3            -1 downto 0);

    signal en_len_a             : std_logic;
    signal en_len_d             : std_logic;
    signal ld_len               : std_logic;
    signal ld_len_orig          : std_logic;
    signal clr_block            : std_logic;
    signal en_block             : std_logic;
    signal fifo_wr_s            : std_logic;
    signal fifo_rd_s            : std_logic;

    type state_type is (S_INIT, S_KEY_CHECK, S_WAIT_KEY_UPDATED,
        S_LD_LEN, S_LD_TAG,
        S_STALL, 
        S_PREP_TAU,                     S_START_TAU_PREP_NPUB,
        S_START_NPUB_PREP_AD,
        S_START_AD_PREP_AD_NEXT,        S_START_AD_LAST,
        S_STORE_HASH,   S_START_PRF,    S_PRF_DONE,

        --! Tiny related states (12)
        S_PREP_TINY1,   S_PREP_TINY2,
        S_PREP_TINY3,   S_PREP_TINY4,
        S_START_TINY,   S_TINY_DONE,
        S_TINY_TAG1,    S_TINY_TAG2,    S_TINY_TAG3,
        S_TINY_TAG4,    S_TINY_TAG5,    S_TINY_TAG6,
        --! Message
        S_PREP_MSG,
        --! Full block
        S_START_CORE_FULL_S,            S_PREP_CORE_FULL,
        S_START_CORE_FULL_LEFT,         S_START_CORE_FULL_RIGHT,
        S_POST_CORE_FULL1,              S_POST_CORE_FULL2,
        --! UV
        S_START_CORE_UV_S,
        S_PREP_CORE_UV,     --! Only for first pass
        S_START_CORE_UV,
        --! XY
        S_PREP_CORE_XY1, S_PREP_CORE_XY2,   --! Only for first pass
        S_PREP_CORE_XY3, S_PREP_CORE_XY4,   --! and there're Mu/Mv
        S_PREP_CORE_XY5,
        S_PREP_CORE_XY6,    --! Only for second pass and there're Mu/Mv
        S_PREP_CORE_XY7,    --! Secondpass: Mu = full block, Mv = empty (special case)
        S_START_CORE_XY,
        S_POST_CORE_XY1, S_POST_CORE_XY2,   --! Post operation if Mu/Mv
        S_POST_CORE_XY3, S_POST_CORE_XY4,
        S_POST_CORE_XY5,
        S_PASS_DONE, 
        
        S_TAG_OUT_STALL, S_TAG_OUT,
        S_ERROR
        );
    signal state, nstate        : state_type;
    signal nnstate_r            : state_type;
    signal nnstate              : state_type;
begin
    p_reg:
    process( clk )
    begin
        if rising_edge( clk ) then
            if rst = '1' then
                state <= S_INIT;
            else
                state <= nstate;
            end if;
            nnstate_r <= nnstate;

            ctrl( 0)            <= en_out       ;
            ctrl( 1)            <= en_ll        ;
            -- ctrl( 2)            <= en_xy or en_xy2;
            ctrl( 3)            <= en_hash      ;
            ctrl( 4)            <= en_ss        ;
            ctrl( 5)            <= en_tmp       ;
            -- ctrl( 6)            <= sel_mask     ;
            ctrl( 7)            <= sel_pad      ;
            ctrl( 8)            <= sel_tau      ;
            --ctrl( 9)            <= sel_ll       ;
            -- 10
            ctrl(13 downto 11)  <= sel_ldata    ;
            ctrl(15 downto 14)  <= sel_ldata_xor;
            ctrl(18 downto 16)  <= sel_data     ;
            ctrl(19)            <= sel_tiny     ;
            ctrl(21 downto 20)  <= sel_xy       ;
            ctrl(22)            <= sel_tag      ;
            ctrl(23)            <= sel_output   ;
            ctrl(24)            <= sel_fifo     ;
            ctrl(25)            <= sel_rothb    ;
            if (is_decrypt = '0') then
                ctrl(28 downto 26)  <= tinyround;
            else
                ctrl(28 downto 26)  <= std_logic_vector(natural(7)
                                       -unsigned(tinyround));
            end if;
            ctrl(29)            <= is_odd       ; --! mask_halfbyte
            ctrl(31 downto 30)  <= sel_rot      ;
            ctrl(32)            <= clr_init     ;
            ctrl(33)            <= sel_sout     ;
            ctrl(34)            <= sel_tmp_xor  ;
            ctrl(35)            <= sel_e        ;
            ctrl(36)            <= sel_otag     ;
            ctrl(37)            <= en_tag       ;

            ctrl_e( 0)          <= estart       ;
            ctrl_e( 1)          <= einit_key    ;
            ctrl_e( 4 downto 2) <= dcalc_mode   ;
            ctrl_e( 7 downto 5) <= dcalc_init   ;
            ctrl_e(15 downto 8) <= block_number ;
            ctrl_e(16)          <= eclr         ;

            rtr                 <= rtr_s        ;
            bdi_ready           <= bdi_ready_s  ;
            bdo_valid           <= bdo_valid_s  ;
            msg_auth_done       <= msg_auth_done_s;

            fdo_valid           <= fifo_wr_s    ;            

            if (clr_block = '1') then
                block_number <= (others => '0');
            elsif (en_block = '1') then
                block_number <= std_logic_vector(unsigned(block_number) + natural(1));
            end if;

            if (state = S_LD_LEN) then
                is_decrypt <= decrypt;
            end if;

            if (ld_len = '1') then
                len_a_r <= len_a;
                len_d_r <= len_d;
                len_d_orig_r <= len_d;
            else
                if (en_len_a = '1') then
                    len_a_r(15 downto 4) <=
                        std_logic_vector(unsigned(len_a_r(15 downto 4))
                        - natural(1));
                end if;
                if (ld_len_orig = '1') then
                    len_d_r <= len_d_orig_r;
                elsif (en_len_d = '1') then
                    len_d_r(15 downto 4) <=
                        std_logic_vector(unsigned(len_d_r(15 downto 4))
                        - natural(1));
                end if;
            end if;

            if (clr_tinyround = '1') then
                tinyround <= (others => '0');
            elsif (en_tinyround = '1') then
                tinyround <= std_logic_vector(unsigned(tinyround) + natural(1));
            end if;

            if (clr_init = '1') then
                is_odd   <= '0';
            elsif (set_odd = '1') then
                is_odd   <= '1';
            end if;

            if (clr_init = '1') then
                is_secondpass <= '0';
            elsif (set_secondpass = '1') then
                is_secondpass <= '1';
            end if;

            if (clr_init = '1' or set_secondpass = '1') then
                is_notfirst <= '0';
            elsif (set_notfirst = '1') then
                is_notfirst <= '1';
            end if;

            if (clr_store_sum = '1') then
                store_sum <= '0';
            elsif (set_store_sum = '1') then
                store_sum <= '1';
            end if;

            if (clr_init = '1' or (save_hash = '1' and edone_r = '1')) then
                save_hash <= '0';
            elsif (set_save_hash_r = '1') then
                save_hash <= '1';
            end if;
            en_xy_r <= en_xy or en_xy2;

            if (clr_pad = '1') then
                is_pad <= '0';
            elsif (set_pad = '1') then
                is_pad <= '1';
            end if;

            if (clr_sel_ll = '1') then
                sel_ll <= '0';
            elsif (set_sel_ll = '1') then
                sel_ll <= '1';
            end if;

            if (clr_init = '1') then
                is_postproc <= '0';
            elsif (set_postproc = '1') then
                is_postproc <= '1';
            end if;

            set_save_hash_r <= set_save_hash;
            edone_r <= edone;
        end if;
    end process;
    en_xy2  <= store_sum and edone;
    en_hash <= save_hash and edone_r;
    --! Not registered outputs
    ctrl( 2)  <= en_xy_r  ;   --! Already registered
    ctrl( 6)  <= sel_mask ;
    ctrl( 9)  <= sel_ll   ;    
    fdi_ready <= fifo_rd_s;

    --! Combinational controls
    dcalc_init <= new_i when sel_i = '1' else block_number(2 downto 0);

    masksize <= std_logic_vector(natural(0) - unsigned(len_d_r(3 downto 0)));    
    with sel_dsize select
    data_size  <= '1' & len_d_r(3 downto 1)                 when "000",
                  -- std_logic_vector(natural(8)
                    -- -unsigned('0' & len_d_r(3 downto 1)))   when "001",
                  '0' & masksize(3 downto 1) when "001",
                  std_logic_vector(
                      unsigned('1' & len_d_r(3 downto 1))
                      -unsigned(len_d_r(3 downto 0)))       when "010",
                  len_a_r(3 downto 0)                       when "011",
                  len_d_r(3 downto 0)                       when "100",
                  std_logic_vector(natural(0)
                    -unsigned(len_d_r(3 downto 0)))         when "101",
                  (others => '1')                           when others;

    with sel_rot_size select
    rtr_s <= len_d_r(3 downto 0)                            when "000",
             '1' & len_d_r(3 downto 1)                      when "001",
             std_logic_vector(natural(0)
                  -unsigned(len_d_r(3 downto 0)))           when "010",
             std_logic_vector(
                  unsigned('1' & len_d_r(3 downto 1))
                  -unsigned(len_d_r(3 downto 0)))           when "011",
             (others => '0')                                when others;


    p_state:
    process( state, bdi_valid, decrypt,
        bdo_ready, key_update, init_key_done,
        len_a_r, len_d_r, len_d_orig_r, eready, edone,
        is_decrypt, tinyround,
        is_secondpass, is_postproc, is_notfirst)
    begin
        --! Default values
        nstate              <= state;
        nnstate             <= nnstate_r;
        --!     External
        bdi_ready_s         <= '0';
        bdo_valid_s         <= '0';
        msg_auth_done_s     <= '0';
        fifo_rd_s           <= '0';
        fifo_wr_s           <= '0';
        --!     Datapath (AEZ_E)
        estart              <= '0';
        einit_key           <= '0';
        dcalc_mode          <= (others => '0');
        eclr                <= '0';
        --!     Top
        en_out              <= '0';
        en_ll               <= '0';
        en_xy               <= '0';
        en_ss               <= '0';
        en_tmp              <= '0';
        sel_mask            <= '0';
        sel_pad             <= '0';
        sel_tau             <= '0';
        sel_rot             <= "00";
        sel_rothb           <= '0';
        sel_ldata           <= (others => '0');
        sel_ldata_xor       <= (others => '0');
        sel_data            <= (others => '0');
        sel_tiny            <= '0';
        sel_xy              <= (others => '0');
        sel_tag             <= '0';
        sel_output          <= '0';
        sel_fifo            <= '0';
        sel_sout            <= '0';
        sel_tmp_xor         <= '0';
        sel_e               <= '0';
        sel_otag            <= '0';
        en_tag              <= '0';
        --! Internal
        clr_init            <= '0';
        clr_block           <= '0';
        sel_i               <= '0';
        new_i               <= (others => '0');
        en_block            <= '0';
        en_len_a            <= '0';
        en_len_d            <= '0';
        ld_len              <= '0';
        ld_len_orig         <= '0';
        set_pad             <= '0';
        clr_pad             <= '0';
        set_store_sum       <= '0';
        clr_store_sum       <= '0';
        set_save_hash       <= '0';
        sel_dsize           <= (others => '0');
        sel_rot_size        <= (others => '0');
        set_odd             <= '0';
        en_tinyround        <= '0';
        clr_tinyround       <= '0';
        set_secondpass      <= '0';
        clr_sel_ll          <= '0';
        set_sel_ll          <= '0';
        set_postproc        <= '0';
        set_notfirst        <= '0';

        case state is
            when S_INIT =>
                nstate          <= S_KEY_CHECK;
                clr_init        <= '1';
                clr_block       <= '1';
                clr_pad         <= '1';
                clr_store_sum   <= '1';
                clr_tinyround   <= '1';
                eclr            <= '1';
                sel_xy          <= "11";
                en_xy           <= '1';
                sel_tmp_xor     <= '1';
                en_tmp          <= '1';
                clr_sel_ll      <= '1';


            when S_KEY_CHECK =>
                if (key_update = '0' and bdi_valid = '1' and eready = '1') then
                    nstate        <= S_LD_LEN;
                elsif (key_update = '1' and eready = '1') then
                    nstate        <= S_WAIT_KEY_UPDATED;
                    einit_key     <= '1';
                end if;
                sel_ldata   <= "010";
                en_ll       <= '1';

            when S_WAIT_KEY_UPDATED =>
                if (init_key_done = '1') then
                    nstate        <= S_LD_LEN;
                end if;

            when S_LD_LEN =>
                bdi_ready_s <= '1';
                if (bdi_valid = '1') then
                    ld_len <= '1';
                    nstate <= S_PREP_TAU;
                end if;
            
            when S_PREP_TAU =>
                nnstate       <= S_START_TAU_PREP_NPUB;
                if (is_decrypt = '0') then
                    nstate    <= S_STALL;
                else
                    nstate    <= S_LD_TAG;
                end if;
                estart        <= '1';
                dcalc_mode    <= std_logic_vector(to_unsigned(3, dcalc_mode'length));

            when S_STALL =>
                nstate <= nnstate_r;
                
            when S_LD_TAG =>
                if (bdi_valid = '1') then
                    bdi_ready_s <= '1';
                    en_tag      <= '1';
                    nstate      <= S_START_TAU_PREP_NPUB;
                end if;

            when S_START_TAU_PREP_NPUB =>
                --! Select Tau
                sel_data      <= "010";
                sel_tau       <= '1';
                --! Prep Npub
                dcalc_mode    <= std_logic_vector(to_unsigned(4, dcalc_mode'length));
                nnstate       <= S_START_NPUB_PREP_AD;
                if ((eready = '1' or edone = '1') and bdi_valid = '1') then
                    nstate        <= S_STALL;
                    --! Start
                    estart        <= '1';
                    --! Load len (Len data should be aviailable by now)
                    set_store_sum <= '1';
                end if;

            when S_START_NPUB_PREP_AD =>
                --! Select Npub
                sel_data      <= "011";
                if (unsigned(len_a_r) >= natural(16)) then
                    dcalc_mode    <= std_logic_vector(to_unsigned(6, dcalc_mode'length));
                    nnstate       <= S_START_AD_PREP_AD_NEXT;
                else
                    dcalc_mode    <= std_logic_vector(to_unsigned(5, dcalc_mode'length));
                    nnstate       <= S_START_AD_LAST;
                end if;
                if (eready = '1' or edone = '1') then
                    nstate        <= S_STALL;
                    bdi_ready_s   <= '1';
                    --! Start
                    estart        <= '1';
                    if (unsigned(len_a_r) > natural(16)) then
                        --! Increment block and decrement len_a
                        en_block      <= '1';
                        en_len_a      <= '1';
                    else
                        set_pad <= '1';
                    end if;
                end if;

            when S_START_AD_PREP_AD_NEXT =>
                --! Prep Next block
                new_i          <= std_logic_vector(to_unsigned(0, new_i'length));
                if (unsigned(len_a_r) >= natural(16)) then
                    dcalc_mode <= std_logic_vector(to_unsigned(6, dcalc_mode'length));
                else
                    sel_i      <= '1';
                    dcalc_mode <= std_logic_vector(to_unsigned(5, dcalc_mode'length));
                end if;
                if (unsigned(len_a_r) > natural(16)) then
                    nnstate    <= S_START_AD_PREP_AD_NEXT;
                else
                    nnstate    <= S_START_AD_LAST;
                end if;
                if ((eready = '1' or edone = '1') and bdi_valid = '1') then
                    if (unsigned(len_a_r) <= natural(16)) then
                        clr_block <= '1';
                        if (unsigned(len_a_r(3 downto 0)) > natural(0)) then
                            set_pad <= '1';
                        end if;
                    end if;
                    nstate    <= S_STALL;
                    --! Start
                    estart        <= '1';
                    --! Read data
                    bdi_ready_s    <= '1';
                    --! Increment block and decrement len_a
                    en_block      <= '1';
                    en_len_a      <= '1';
                end if;

            when S_START_AD_LAST =>
                sel_i <= '1';
                if (unsigned(len_d_r) = natural(0)) then
                    --! Prep PRF
                    dcalc_mode    <= std_logic_vector(to_unsigned(7, dcalc_mode'length));
                    new_i         <= std_logic_vector(to_unsigned(3, new_i'length));
                    --! PRF
                    nnstate       <= S_STORE_HASH;
                else
                    --! Prep Tiny
                    dcalc_mode    <= std_logic_vector(to_unsigned(0, dcalc_mode'length));
                    new_i         <= std_logic_vector(to_unsigned(6, new_i'length));
                    if (unsigned(len_d_r) >= natural(16)) then
                        --! Msg
                        nnstate   <= S_PREP_MSG;
                    else
                        --! Tiny
                        nnstate   <= S_PREP_TINY1;
                    end if;
                end if;

                if ((eready = '1' or edone = '1') and bdi_valid = '1') then
                    set_save_hash <= '1';
                    estart      <= '1';
                    bdi_ready_s  <= '1';
                    clr_block   <= '1';
                    clr_pad     <= '1';
                    if (unsigned(len_d_r) < natural(16)) then
                        set_odd     <= len_d_r(0);
                    end if;
                    nstate      <= S_STALL;
                end if;
                sel_pad     <= is_pad;
                sel_mask    <= is_pad;
                sel_dsize   <= "011";

            when S_STORE_HASH =>
                if (edone = '1') then
                    nstate <= S_START_PRF;
                end if;

            when S_START_PRF =>
                sel_data    <= "110";
                nnstate     <= S_PRF_DONE;
                nstate      <= S_STALL;
                estart      <= '1';

            when S_PRF_DONE =>
                sel_tag  <= '1';
                sel_otag <= '1';
                if (eready = '1' or edone = '1') then
                    if (is_decrypt = '1') then                        
                        nstate          <= S_INIT;
                        msg_auth_done_s <= '1';
                    elsif (bdo_ready = '1') then
                        --! Tag write
                        nstate          <= S_INIT;
                        bdo_valid_s     <= '1';
                    end if;
                end if;

            when S_PREP_TINY1 =>
                --! Step 1) Get tag
                --! Step 2) Mask the top part of tag
                --! Step 3) Rotate the top part to the end of message
                --!         and combine it with data
                --! Step 4) ll_r = tmp_r
                --!         Get Tag + Rotate                
                if (is_decrypt = '1') then
                    sel_rot <= "11";
                end if;
                sel_rot_size<= "111";
                sel_tmp_xor <= '1';
                en_tmp      <= '1';
                nstate      <= S_PREP_TINY2;

            when S_PREP_TINY2 =>
                --! Mask the top part of tag
                nstate      <= S_PREP_TINY3;
                sel_data    <= "100";
                sel_mask    <= '1';
                sel_dsize   <= "001";
                en_tmp      <= '1';

            when S_PREP_TINY3 =>
                sel_data    <= "000";
                sel_rot     <= "10";
                sel_rot_size<= "010";
                if (bdi_valid = '1') then
                    nstate      <= S_PREP_TINY4;
                    bdi_ready_s  <= '1';
                    en_tmp      <= '1';
                end if;

            when S_PREP_TINY4 =>
                --! ll_r = tmp_r
                sel_ldata   <= "010";
                en_ll       <= '1';
                --! tmp_r = rot(tag)
                sel_tmp_xor <= '1';
                if (is_decrypt = '1') then
                    sel_rot     <= "11";
                end if;
                sel_rot_size<= "011";
                sel_rothb   <= is_odd;
                en_tmp      <= '1';
                nstate      <= S_START_TINY;

            when S_START_TINY =>
                clr_store_sum <= '1';
                sel_i       <= '1';
                dcalc_mode  <= std_logic_vector(to_unsigned(0, dcalc_mode'length));
                new_i       <= std_logic_vector(to_unsigned(6, new_i'length));
                sel_tiny    <= '1';
                if (unsigned(tinyround) = natural(0)) then
                    sel_data <= "100";
                else
                    sel_data <= "111";
                end if;
                sel_mask    <= '1';
                sel_pad     <= '1';
                sel_dsize   <= "000";
                sel_ldata   <= "010";
                if (unsigned(tinyround) = natural(7)) then
                    nnstate     <= S_TINY_DONE;
                else
                    nnstate     <= S_START_TINY;
                end if;
                if (eready = '1' or edone = '1') then
                    estart       <= '1';
                    en_tinyround <= '1';
                    nstate       <= S_STALL;
                    if (unsigned(tinyround) /= natural(0)) then
                        en_tmp   <= '1';
                        en_ll    <= '1';
                    end if;
                end if;

            when S_TINY_DONE =>
                --! Store final data
                sel_ldata   <= "010";
                --! Remove padding bit
                sel_data    <= "111";
                sel_mask    <= '1';
                sel_dsize   <= "000";
                if (eready = '1' or edone = '1') then
                    --! Write output
                    bdo_valid_s <= '1';
                    en_ll       <= '1';
                    en_tmp      <= '1';
                    --! Needs to perform the following steps to
                    --! get the right tag output
                    --! Step 1) Rotate tmp_r (left part)
                    --! Step 2) Mask tmp_r to remove ciphertext
                    --! Step 3) Switch ll_r and tmp_r
                    --!         At the same time, mask ll_r
                    --! Step 4) Rotate tmp_r
                    --! Step 5) Combine
                    --! Step 6) Output Tag
                    nstate      <= S_TINY_TAG1;
                end if;

            when S_TINY_TAG1 =>
                --! (Step 1)
                nstate  <= S_TINY_TAG2;
                --! Rotate data out of tag
                sel_tmp_xor  <= '1';
                sel_rot      <= "10";
                sel_rot_size <= "000";
                en_tmp       <= '1';
                --! Clear AEZ_E
                eclr        <= '1';

            when S_TINY_TAG2 =>
                --! (Step 2)
                nstate      <= S_TINY_TAG3;
                sel_data    <= "100";
                sel_dsize   <= "010";
                sel_mask    <= '1';
                en_tmp      <= '1';


            when S_TINY_TAG3 =>
                --! (Step 3)
                nstate      <= S_TINY_TAG4;
                --! Store tmp_r in ll_r
                sel_ldata   <= "010";
                en_ll       <= '1';
                --! Store ll_r in tmp_r
                sel_data    <= "111";
                en_tmp      <= '1';
                --! remove padding data
                sel_dsize   <= "000";
                sel_mask    <= '1';

            when S_TINY_TAG4 =>
                --! (Step 4)
                nstate      <= S_TINY_TAG5;
                sel_tmp_xor <= '1';
                sel_rot     <= "10";
                sel_rot_size<= "001";
                sel_rothb   <= is_odd;
                en_tmp      <= '1';

            when S_TINY_TAG5 =>
                --! (Step 5)
                nstate      <= S_TINY_TAG6;
                --! Combine shifted L (tmp_r) and shifted R (ll_r)
                sel_data    <= "111";
                sel_rot     <= "10";
                sel_rot_size<= "111"; --! Dont rotate
                en_tmp      <= '1';


            when S_TINY_TAG6 =>
                --! (Step 6)
                sel_otag <= '1';
                if (is_decrypt = '1') then
                    msg_auth_done_s <= '1';
                    nstate          <= S_INIT;
                elsif (bdo_ready = '1') then
                    bdo_valid_s     <= '1';
                    nstate          <= S_INIT;
                end if;

            when S_PREP_MSG =>
                --! Prep Delta for the first AEZ_CORE operation
                sel_i         <= '1';
                if (unsigned(len_d_r) >= natural(48)) then
                    --! Prep M1
                    if (is_secondpass = '0') then
                        dcalc_mode    <= std_logic_vector(to_unsigned(1, dcalc_mode'length));
                        nnstate       <= S_START_CORE_FULL_RIGHT;
                        sel_data      <= "000";
                    else
                        sel_data      <= "001";
                        sel_ldata     <= "001";
                        dcalc_mode    <= std_logic_vector(to_unsigned(2, dcalc_mode'length));
                        nnstate       <= S_START_CORE_FULL_S;
                    end if;
                    new_i         <= std_logic_vector(to_unsigned(0, new_i'length));

                elsif (unsigned(len_d_r) > natural(16)) then
                    --! Prep Mu
                    if (is_secondpass = '0') then
                        dcalc_mode    <= std_logic_vector(to_unsigned(0, dcalc_mode'length));
                        nnstate       <= S_PREP_CORE_UV;
                    else
                        sel_ldata     <= "001";
                        dcalc_mode    <= std_logic_vector(to_unsigned(7, dcalc_mode'length));
                        nnstate       <= S_START_CORE_UV_S;
                    end if;
                    new_i       <= std_logic_vector(to_unsigned(4, new_i'length));
                    sel_tmp_xor <= '1';
                else
                    --! Prep Mx
                    if (is_secondpass = '0') then
                        dcalc_mode    <= std_logic_vector(to_unsigned(0, dcalc_mode'length));
                    else
                        dcalc_mode    <= std_logic_vector(to_unsigned(7, dcalc_mode'length));
                    end if;
                    if (is_decrypt = '1' and is_secondpass = '0')
                        or (is_decrypt = '0' and is_secondpass = '1')
                    then
                        new_i         <= std_logic_vector(to_unsigned(2, new_i'length));
                    else
                        new_i         <= std_logic_vector(to_unsigned(1, new_i'length));
                    end if;
                    --! Mx
                    nnstate         <= S_START_CORE_XY;
                    --! Select Hash xor X/Y
                    if (is_secondpass = '1') then
                        sel_ldata       <= "001";
                    else
                        sel_ldata_xor   <= "01";
                    end if;
                    if (is_decrypt = '1') then
                        sel_rot <= "11";
                    end if;
                    sel_tmp_xor <= '1';
                end if;
                sel_rot_size  <= "111";

                if ((eready = '1')
                    and (bdi_valid = '1' or is_secondpass = '1'))
                then
                    nstate          <= S_STALL;
                    clr_store_sum   <= '1';

                    if (unsigned(len_d_r) >= natural(48)) then
                        --! Increment block and decrement len_d
                        en_len_d    <= '1';                        
                        set_sel_ll  <= '1';
                        if (is_secondpass = '0') then
                            bdi_ready_s  <= '1';
                        else
                            fifo_rd_s   <= '1';
                        end if;
                    elsif (unsigned(len_d_r) > natural(16)) then
                        en_len_d    <= '1';
                        if (is_secondpass = '1') then
                            fifo_rd_s   <= '1';
                        end if;
                    else
                        --! Read data
                        if (is_secondpass = '0') then
                            bdi_ready_s  <= '1';
                        else
                            fifo_rd_s   <= '1';
                        end if;
                    end if;
                    en_tmp      <= '1';
                    en_ll       <= '1';
                    --! Start
                    estart      <= '1';
                end if;

            when S_START_CORE_FULL_S =>         
                if (is_notfirst = '1') then
                    nnstate     <= S_PREP_CORE_FULL;
                else
                    nnstate     <= S_START_CORE_FULL_RIGHT;
                end if;
                dcalc_mode  <= std_logic_vector(to_unsigned(0, dcalc_mode'length));
                new_i       <= std_logic_vector(to_unsigned(0, new_i'length));
                sel_i       <= '1';
                sel_data    <= "101";

                if (edone = '1' or eready = '1')
                    and (is_notfirst = '0' or bdo_ready = '1')
                then
                    estart          <= '1';
                    set_store_sum   <= '1';                    
                    if (is_notfirst = '1') then
                        bdo_valid_s <= '1';
                    end if;
                    nstate <= S_STALL;
                end if;
                
            when S_PREP_CORE_FULL =>
                --! Special state to load data from FIFO onto tmp_r.
                --! This state is operating in parallel with 
                --! S_START_CORE_FULL_S so it is considered as free.
                --! Note: No need to reissue the fifo_rd_s signal, it has 
                --!       already been issued in S_START_CORE_FULL_LEFT                
                
                nstate      <= S_START_CORE_FULL_RIGHT;
                sel_data    <= "001";
                en_tmp      <= '1';
                sel_ldata   <= "001";
                en_ll       <= '1';
                
                

            when S_START_CORE_FULL_RIGHT =>
                if (is_secondpass = '0') then
                    dcalc_mode  <= std_logic_vector(to_unsigned(0, dcalc_mode'length));
                    new_i       <= std_logic_vector(to_unsigned(0, new_i'length));
                    sel_i       <= '1';
                else
                    dcalc_mode  <= std_logic_vector(to_unsigned(1, dcalc_mode'length));
                end if;

                --! o Starts processing from bdi/fifo
                --! o stores the processing data onto tmp_r and xor it with E
                --! o Copy left data (tmp_r) onto ll_r
                sel_ldata   <= "010";
                if (is_secondpass = '1') then
                    sel_data <= "001";
                    sel_sout <= '1';
                    sel_ldata_xor <= "10";
                    sel_e         <= '1';
                end if;
                sel_rot_size    <= "111";
                sel_output      <= '1';
                nnstate         <= S_START_CORE_FULL_LEFT;

                if (edone = '1' or eready = '1')
                    and ((is_secondpass = '1' and bdo_ready = '1')
                        or (is_secondpass = '0' and bdi_valid = '1'))
                then
                    clr_store_sum <= '1';
                    if (is_secondpass = '0' and bdi_valid = '1') then
                        bdi_ready_s <= '1';
                        if (is_notfirst = '1') then
                            fifo_wr_s     <= '1';
                        end if;
                    elsif (is_secondpass = '1') then
                        fifo_rd_s  <= '1';
                        if (is_notfirst = '1') then
                            bdo_valid_s <= '1';
                        end if;
                    end if;
                    en_ll       <= '1';
                    en_tmp      <= '1';
                    en_len_d    <= '1';
                    en_block    <= '1';
                    estart      <= '1';                    
                    nstate      <= S_STALL;
                end if;

            when S_START_CORE_FULL_LEFT =>
                if (unsigned(len_d_r) >= natural(48)) then
                    --! Prep M1
                    if (is_secondpass = '0') then
                        dcalc_mode    <= std_logic_vector(to_unsigned(1, dcalc_mode'length));
                        nnstate       <= S_START_CORE_FULL_RIGHT;
                    else
                        dcalc_mode    <= std_logic_vector(to_unsigned(2, dcalc_mode'length));
                        nnstate       <= S_START_CORE_FULL_S;
                    end if;
                elsif (unsigned(len_d_r) > natural(16)) then
                    --! Prep Mu
                    if (is_secondpass = '0') then
                        dcalc_mode    <= std_logic_vector(to_unsigned(0, dcalc_mode'length));
                    else
                        dcalc_mode    <= std_logic_vector(to_unsigned(7, dcalc_mode'length));
                    end if;
                    new_i         <= std_logic_vector(to_unsigned(4, new_i'length));
                    sel_i         <= '1';
                    nnstate       <= S_POST_CORE_FULL1;
                else
                    --! Prep Mx
                    if (is_secondpass = '0') then
                        dcalc_mode    <= std_logic_vector(to_unsigned(0, dcalc_mode'length));
                    else
                        dcalc_mode    <= std_logic_vector(to_unsigned(7, dcalc_mode'length));
                    end if;
                    if (is_decrypt = '1' and is_secondpass = '0')
                        or (is_decrypt = '0' and is_secondpass = '1')
                    then
                        new_i         <= std_logic_vector(to_unsigned(2, new_i'length));
                    else
                        new_i         <= std_logic_vector(to_unsigned(1, new_i'length));
                    end if;
                    sel_i         <= '1';
                    nnstate       <= S_POST_CORE_FULL1;
                end if;

                --! o Starts processing from chain value (xor_main)
                --! o Load next data block onto tmp_r
                --! o Copy right data (tmp_r) onto ll_r
                --! o store output in out_r
                sel_ldata       <= "010";
                sel_data        <= "111";
                sel_tmp_xor     <= '1';
                sel_rot         <= "01";
                sel_rot_size    <= "111";

                if (edone = '1' or eready = '1')
                    and (bdi_valid = '1' or is_secondpass = '1')
                then
                    en_out      <= '1';
                    if (unsigned(len_d_r) >= natural(48)) then
                        if (is_secondpass = '0') then
                            bdi_ready_s  <= '1';
                        else
                            fifo_rd_s   <= '1';
                        end if;
                        en_len_d    <= '1';                        
                    end if;
                    if (is_secondpass = '0') then
                        set_store_sum <= '1';
                        fifo_wr_s     <= '1';
                    else
                        clr_store_sum <= '1';
                    end if;
                    set_notfirst  <= '1';
                    en_ll       <= '1';
                    en_tmp      <= '1';
                    estart      <= '1';
                    nstate      <= S_STALL;
                end if;

            when S_POST_CORE_FULL1 =>
                nnstate <= S_POST_CORE_FULL2;
                if (edone = '1' or eready = '1')
                    and (bdo_ready = '1' or is_secondpass = '0')
                then
                    clr_store_sum   <= '1';
                    if (is_secondpass = '1') then
                        bdo_valid_s     <= '1';
                    else
                        fifo_wr_s       <= '1';
                    end if;
                    nstate          <= S_STALL;
                end if;

            when S_POST_CORE_FULL2 =>
                clr_sel_ll  <= '1';
                sel_output  <= '1';
                if (unsigned(len_d_r) > natural(16)) then
                    if (is_secondpass = '1') then
                        sel_ldata     <= "001";
                    end if;
                else
                    --! Select Hash xor X/Y
                    if (is_secondpass = '1') then
                        sel_ldata       <= "001";
                    else
                        sel_ldata_xor   <= "01";
                    end if;
                    if (is_decrypt = '1') then
                        sel_rot <= "11";
                    end if;
                end if;
                sel_rot_size  <= "111";
                sel_tmp_xor <= '1';

                if (bdi_valid = '1' or is_secondpass = '1')
                    and (bdo_ready = '1')
                then
                    if (unsigned(len_d_r) > natural(16)) then
                        en_len_d    <= '1';
                        eclr        <= '1';
                        if (is_secondpass = '1') then
                            fifo_rd_s   <= '1';
                            nstate <= S_START_CORE_UV_S;
                        else
                            nstate <= S_PREP_CORE_UV;
                        end if;
                    else
                        --! Read data
                        if (is_secondpass = '0') then
                            bdi_ready_s  <= '1';
                        else
                            fifo_rd_s   <= '1';
                        end if;
                        nstate <= S_START_CORE_XY;
                    end if;
                    en_tmp      <= '1';
                    en_ll       <= '1';
                    if (is_secondpass = '1') then
                        bdo_valid_s     <= '1';
                    end if;
                end if;


            when S_START_CORE_UV_S =>
                --! Prep next
                sel_data <= "101";
                sel_i    <= '1';

                dcalc_mode    <= std_logic_vector(to_unsigned(0, dcalc_mode'length));
                if (tinyround(0) = '0') then
                    new_i         <= std_logic_vector(to_unsigned(4, new_i'length));
                else
                    new_i         <= std_logic_vector(to_unsigned(5, new_i'length));
                end if;
                nnstate   <= S_START_CORE_UV;
                if (eready = '1' or edone = '1') then
                    estart          <= '1';
                    clr_store_sum   <= '1';
                    nstate          <= S_STALL;
                end if;

            when S_PREP_CORE_UV =>
                if (edone = '1' or eready = '1')
                    and ((is_secondpass = '1')
                        or (is_secondpass = '0'
                            and bdi_valid = '1')
                        or (unsigned(len_d_r) = natural(0)))
                then
                    eclr    <= '1';
                    if (is_secondpass = '0'
                        and bdi_valid = '1'
                        and unsigned(len_d_r) /= natural(0))
                    then
                        en_ll      <= '1';
                    end if;
                    nstate  <= S_START_CORE_UV;
                end if;

            when S_START_CORE_UV =>
                sel_i           <= '1';
                if (unsigned(len_d_r) >= natural(16)) then
                    if (is_secondpass = '0') then
                        dcalc_mode  <= std_logic_vector(to_unsigned(0, dcalc_mode'length));
                        new_i       <= std_logic_vector(to_unsigned(5, new_i'length));
                        nnstate     <= S_PREP_CORE_UV;
                    else
                        dcalc_mode  <= std_logic_vector(to_unsigned(7, dcalc_mode'length));
                        new_i       <= std_logic_vector(to_unsigned(5, new_i'length));
                        nnstate     <= S_START_CORE_UV_S;
                    end if;
                else
                    if (is_secondpass = '0') then
                        dcalc_mode  <= std_logic_vector(to_unsigned(0, dcalc_mode'length));
                        if (is_decrypt = '0') then
                            new_i       <= std_logic_vector(to_unsigned(1, new_i'length));
                        else
                            new_i       <= std_logic_vector(to_unsigned(2, new_i'length));
                        end if;
                    else
                        dcalc_mode  <= std_logic_vector(to_unsigned(7, dcalc_mode'length));
                        if (is_decrypt = '0') then
                            new_i       <= std_logic_vector(to_unsigned(2, new_i'length));
                        else
                            new_i       <= std_logic_vector(to_unsigned(1, new_i'length));
                        end if;
                    end if;
                    sel_dsize   <= "100";
                    sel_mask    <= '1';
                    sel_pad     <= '1';

                    if (unsigned(len_d_r) /= natural(0)) then
                        set_postproc    <= '1';
                        if (is_secondpass = '0') then
                            nnstate     <= S_PREP_CORE_XY1;
                        else
                            nnstate     <= S_PREP_CORE_XY6;
                        end if;
                    else
                        nnstate <= S_PREP_CORE_XY7;
                    end if;
                end if;
                if (is_secondpass = '1') then
                    sel_ldata       <= "001";
                    sel_data        <= "111";
                end if;
                --! Rotate input and store it in tmp_r
                sel_rot         <= "01";
                sel_rot_size    <= "000";
                sel_tmp_xor     <= '1';

                if ((eready = '1' or edone = '1')
                    and ((bdi_valid = '1' and is_secondpass = '0')
                        or (is_secondpass = '1' and bdo_ready = '1')))
                then
                    set_store_sum   <= '1';
                    if (unsigned(len_d_r) >= natural(16)) then
                        en_len_d        <= '1';
                        en_tinyround    <= '1';
                    else
                        clr_tinyround   <= '1';
                    end if;
                    --! Start
                    estart      <= '1';
                    en_tmp      <= '1';
                    en_ll       <= '1';
                    nstate      <= S_STALL;
                    --! Read data and write to FIFO
                    if (is_secondpass = '0') then
                        if (unsigned(len_d_r) /= natural(0)) then
                            bdi_ready_s  <= '1';
                        end if;
                        fifo_wr_s   <= '1';
                    else
                        fifo_rd_s   <= '1';
                        if (unsigned(len_d_r) >= natural(16)) then
                            bdo_valid_s <= '1';
                        elsif (unsigned(len_d_r) /= natural(0)) then
                            fifo_wr_s   <= '1';
                        end if;
                        --! Note: Ciphertext/Plaintext is written if it is
                        --!       a full block. Output is written back onto
                        --!       the FIFO if it's a partial block for
                        --!       further processing.
                    end if;
                end if;

            when S_PREP_CORE_XY1 =>
                --! Step 1: Mask the used data bit (data that have been used in S_START_CORE_UV)
                --! Step 2: Rotate tmp_r for 0-len_d_r
                --!         then combine data with the last block.
                --! Step 3: Rotate again to put data at the right position
                --! Step 4: Wait until last X/Y is computed and stored Xu/Xv in xy_r
                --! Step 5: Load data onto ll_r and tag/exp_tag onto tmp_r
                sel_mask <= '1';
                sel_data <= "100";
                sel_dsize<= "101";
                en_tmp   <= '1';
                nstate   <= S_PREP_CORE_XY2;


            when S_PREP_CORE_XY2 =>
                sel_rot         <= "10";
                sel_rot_size    <= "010";
                sel_dsize       <= "101";
                if (bdi_valid = '1') then
                    bdi_ready_s <= '1';
                    en_tmp      <= '1';
                    nstate      <= S_PREP_CORE_XY3;
                end if;

            when S_PREP_CORE_XY3 =>
                sel_rot         <= "10";
                sel_rot_size    <= "000";
                sel_tmp_xor     <= '1';
                en_tmp          <= '1';
                nstate          <= S_PREP_CORE_XY4;

            when S_PREP_CORE_XY4 =>
                if (edone = '1' or eready = '1') then
                    nstate          <= S_PREP_CORE_XY5;
                end if;

            when S_PREP_CORE_XY5 =>
                sel_tmp_xor     <= '1';
                sel_ldata       <= "010";
                sel_ldata_xor   <= "01";
                if (is_decrypt = '1') then
                    sel_rot <= "11";
                end if;
                sel_rot_size    <= "111";   --! No rotation
                en_ll           <= '1';
                en_tmp          <= '1';
                nstate          <= S_START_CORE_XY;

            when S_PREP_CORE_XY6 =>
                --! Second pass: Needs to wait until the last block is done.
                if (eready = '1') then
                    nstate      <= S_START_CORE_XY;
                end if;

            when S_PREP_CORE_XY7 =>
                --! Second pass: In the case that Mv is empty (Mu = 16 bytes)
                --!         load data onto ll_r
                --!         load tag onto tmp_r
                if (eready = '1')
                    and (bdi_valid = '1' or is_secondpass = '1')
                then
                    nstate      <= S_START_CORE_XY;
                    if (is_secondpass = '0') then
                        bdi_ready_s  <= '1';
                    end if;
                    en_ll       <= '1';
                    en_tmp      <= '1';
                end if;
                if (is_secondpass = '1') then
                    sel_ldata   <= "001";
                else
                    sel_ldata_xor <= "01";
                end if;

                if (is_decrypt = '1') then
                    sel_rot     <= "11";
                end if;
                sel_rot_size <= "111";
                sel_tmp_xor  <= '1';

            when S_START_CORE_XY =>
                clr_store_sum  <= '1';
                if (is_secondpass = '0' and tinyround(0) = '0') then
                    dcalc_mode    <= std_logic_vector(to_unsigned(7, dcalc_mode'length));
                else
                    dcalc_mode    <= std_logic_vector(to_unsigned(0, dcalc_mode'length));
                end if;
                sel_i    <= '1';
                if (is_decrypt = '1' and is_secondpass = '0')
                    or (is_decrypt = '0' and is_secondpass = '1')
                then
                    new_i         <= std_logic_vector(to_unsigned(2, new_i'length));
                else
                    new_i         <= std_logic_vector(to_unsigned(1, new_i'length));
                end if;
                if (tinyround(0) = '1') then
                    nnstate <= S_PASS_DONE;
                else
                    nnstate <= S_START_CORE_XY;
                end if;
                if (tinyround(0) = '0') then
                    if (is_secondpass = '0') then
                        sel_data    <= "100";
                    else
                        sel_data    <= "001";
                    end if;
                else
                    sel_data  <= "111";
                    sel_ldata <= "010";
                end if;
                if (tinyround(0) = '1' and is_secondpass = '1') then
                    sel_ldata_xor   <= "01";
                end if;


                if (eready = '1' or edone = '1') then
                    nstate      <= S_STALL;
                    estart      <= '1';
                    if (tinyround(0) = '0' and is_secondpass = '1') then
                        fifo_rd_s   <= '1';
                    end if;
                    if (tinyround(0) = '1' and is_secondpass = '0') then
                        fifo_wr_s   <= '1';
                    end if;
                    if (tinyround(0) = '1') then
                        clr_tinyround   <= '1';
                    else
                        en_tinyround    <= '1';
                    end if;
                    if (tinyround(0) = '1') then
                        en_ll   <= '1';
                        en_ss   <= '1';
                    end if;
                    en_tmp      <= '1';
                end if;

            when S_PASS_DONE =>                
                if (is_secondpass = '0') then
                    --! xy_r gets hash
                    sel_xy      <= "01";
                else
                    --! xy_r gets tag
                    sel_xy      <= "10";
                end if;
                --! Second pass related controls
                ld_len_orig <= '1';
                clr_block   <= '1';
                sel_data    <= "111";
                sel_dsize   <= "101";
                sel_mask    <= '1';
                sel_ldata   <= "011";
                if ((eready = '1' or edone = '1') 
                        and ((bdo_ready = '1' and is_decrypt = '0')
                            or (is_decrypt = '1')))
                    or (edone = '1' and is_secondpass = '0')
                then
                    if (is_secondpass = '0') then
                        set_secondpass  <= '1';
                        fifo_wr_s       <= '1';
                        en_xy           <= '1';
                        en_ss           <= '1';
                        nstate          <= S_PREP_MSG;
                    else
                        if (is_decrypt = '1') then
                            msg_auth_done_s <= '1';
                        else
                            --! Tag store
                            en_xy <= '1';
                        end if;                        
                        if (is_postproc = '1') then
                            --! Note: The value of len_d_r should be
                            --!       equal to len_d_orig_r at this point
                            --! Prepare to shift data for the last block
                            --! so that it produces the right output.
                            --! o Tag is produced/verified at this step
                            --! o Output is stored in tmp_r and ll_r
                            en_tmp          <= '1';
                            en_ll           <= '1';
                            nstate          <= S_POST_CORE_XY1;
                        else
                            bdo_valid_s     <= '1';
                            if (is_decrypt = '1') then
                                nstate      <= S_INIT;
                            else
                                nstate      <= S_TAG_OUT_STALL;
                            end if;
                        end if;
                    end if;
                end if;

            when S_POST_CORE_XY1 =>
                --! Step 1: Rotate tmp_r and combine with data from FIFO
                --!         clear state
                --! Step 2: Store tmp_r in ll_r
                --!         Store output in tmp_r
                --! Step 3: Write first output
                --!         Rotate bottom part
                --! Step 4: Store tmp_r in ll_r
                --! Step 5: Write last output
                fifo_rd_s       <= '1';
                sel_data        <= "001";
                sel_dsize       <= "100";
                sel_mask        <= '1';
                sel_rot         <= "10";
                sel_rot_size    <= "010";
                en_tmp          <= '1';
                eclr            <= '1';
                nstate          <= S_POST_CORE_XY2;

            when S_POST_CORE_XY2 =>
                sel_ldata       <= "010";
                en_ll           <= '1';
                sel_data        <= "111";
                en_tmp          <= '1';
                nstate          <= S_POST_CORE_XY3;

            when S_POST_CORE_XY3 =>
                sel_tmp_xor  <= '1';
                sel_rot      <= "10";
                sel_rot_size <= "010";
                if (bdo_ready = '1') then
                    bdo_valid_s <= '1';
                    en_tmp      <= '1';
                    nstate      <= S_POST_CORE_XY4;
                end if;

            when S_POST_CORE_XY4 =>
                sel_ldata   <= "010";
                en_ll       <= '1';
                nstate      <= S_POST_CORE_XY5;

            when S_POST_CORE_XY5 =>
                if (bdo_ready = '1') then
                    bdo_valid_s <= '1';
                    if (is_decrypt = '1') then
                        nstate      <= S_INIT;
                    else
                        nstate      <= S_TAG_OUT_STALL;
                    end if;
                end if;
                
            when S_TAG_OUT_STALL =>
                --! Stall for one cycle as output write is a delayed control
                nstate <= S_TAG_OUT;
                
            when S_TAG_OUT =>
                sel_otag    <= '1';                 
                sel_output  <= '1';
                if (is_decrypt = '1') then
                    msg_auth_done_s <= '1';
                else
                    --! Tag write
                    bdo_valid_s     <= '1';
                    if (bdo_ready = '1') then
                        nstate <= S_INIT;
                    end if;
                end if;


            when others =>

        end case;
    end process;
end behavior;

